/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 * 
 * Copyright (©) 2005,2006,2007,2008 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2001,2007 Mark Wedel
 * Copyright (©) 1992,2007 Frank Tore Johansen
 * 
 * Deliantra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

/**
 * \file
 * Client handling.
 *
 * \date 2003-12-02
 *
 * This file implements all of the goo on the server side for handling 
 * clients.  It's got a bunch of global variables for keeping track of 
 * each of the clients. 
 *
 * Note:  All functions that are used to process data from the client
 * have the prototype of (char *data, int datalen, int client_num).  This
 * way, we can use one dispatch table.
 *
 * esrv_map_new starts updating the map
 *
 */

#include <global.h>
#include <sproto.h>

#include <living.h>
#include <commands.h>

/* This block is basically taken from socket.c - I assume if it works there,
 * it should work here.
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <unistd.h>
#include <sys/time.h>

#include "sounds.h"

/**
 * This table translates the attack numbers as used within the
 * program to the value we use when sending STATS command to the
 * client.  If a value is -1, then we don't send that to the
 * client.
 */
short atnr_cs_stat[NROFATTACKS] = { CS_STAT_RES_PHYS,
  CS_STAT_RES_MAG, CS_STAT_RES_FIRE, CS_STAT_RES_ELEC,
  CS_STAT_RES_COLD, CS_STAT_RES_CONF, CS_STAT_RES_ACID,
  CS_STAT_RES_DRAIN, -1 /* weaponmagic */ ,
  CS_STAT_RES_GHOSTHIT, CS_STAT_RES_POISON,
  CS_STAT_RES_SLOW, CS_STAT_RES_PARA, CS_STAT_TURN_UNDEAD,
  CS_STAT_RES_FEAR, -1 /* Cancellation */ ,
  CS_STAT_RES_DEPLETE, CS_STAT_RES_DEATH,
  -1 /* Chaos */ , -1 /* Counterspell */ ,
  -1 /* Godpower */ , CS_STAT_RES_HOLYWORD,
  CS_STAT_RES_BLIND,
  -1,                           /* Internal */
  -1,                           /* life stealing */
  -1                            /* Disease - not fully done yet */
};

static void
socket_map_scroll (client *ns, int dx, int dy)
{
  struct Map newmap;
  int x, y, mx, my;

  ns->send_packet_printf ("map_scroll %d %d", dx, dy);

  /* If we are using the Map1aCmd, we may in fact send
   * head information that is outside the viewable map.
   * So set the mx,my to the max value we want to
   * look for.  Removed code to do so - it caused extra
   * complexities for the client, and probably doesn't make
   * that much difference in bandwidth.
   */
  mx = ns->mapx;
  my = ns->mapy;

  /* the x and y here are coordinates for the new map, i.e. if we moved
   * (dx,dy), newmap[x][y] = oldmap[x-dx][y-dy]. For this reason,
   * if the destination x or y coordinate is outside the viewable
   * area, we clear the values - otherwise, the old values
   * are preserved, and the check_head thinks it needs to clear them.
   */
  for (x = 0; x < mx; x++)
    for (y = 0; y < my; y++)
      if (x >= ns->mapx || y >= ns->mapy)
        /* clear cells outside the viewable area */
        memset (&newmap.cells[x][y], 0, sizeof (struct MapCell));
      else if ((x + dx) < 0 || (x + dx) >= ns->mapx || (y + dy) < 0 || (y + dy) >= ns->mapy)
        /* clear newly visible tiles within the viewable area */
        memset (&(newmap.cells[x][y]), 0, sizeof (struct MapCell));
      else
        memcpy (&(newmap.cells[x][y]), &(ns->lastmap.cells[x + dx][y + dy]), sizeof (struct MapCell));

  memcpy (&(ns->lastmap), &newmap, sizeof (struct Map));

  /* Make sure that the next "map1" command will be sent (even if it is
   * empty).
   */
  ns->sent_scroll = 1;
}

static void
clear_map (player *pl)
{
  pl->ns->mapinfo_queue_clear ();

  memset (&pl->ns->lastmap, 0, sizeof (pl->ns->lastmap));

  pl->ns->force_newmap = false;

  if (pl->ns->newmapcmd == 1)
    pl->ns->send_packet ("newmap");

  pl->ns->floorbox_reset ();
}

static void
send_map_info (player *pl)
{
  client &socket = *pl->ns;
  object *ob = pl->observe;

  if (socket.mapinfocmd)
    {
      if (ob->map && ob->map->path[0])
        {
          int flags = 0;

          if (ob->map->tile_path[0]) flags |= 1;
          if (ob->map->tile_path[1]) flags |= 2;
          if (ob->map->tile_path[2]) flags |= 4;
          if (ob->map->tile_path[3]) flags |= 8;

          socket.send_packet_printf ("mapinfo - spatial %d %d %d %d %d %s",
                                     flags, socket.mapx / 2 - ob->x, socket.mapy / 2 - ob->y,
                                     ob->map->width, ob->map->height, &ob->map->path);
        }
      else
        socket.send_packet ("mapinfo current");
    }
}

/** check for map/region change and send new map data */
static void
check_map_change (player *pl)
{
  client &socket = *pl->ns;
  object *ob = pl->observe;

  region *reg = ob->region ();
  if (socket.current_region != reg)
    {
      INVOKE_PLAYER (REGION_CHANGE, pl, ARG_REGION (reg), ARG_REGION (socket.current_region));
      socket.current_region = reg;
    }

  // first try to aovid a full newmap on tiled map scrolls
  if (socket.current_map != ob->map && !socket.force_newmap)
    {
      rv_vector rv;

      get_rangevector_from_mapcoord (socket.current_map, socket.current_x, socket.current_y, ob, &rv, 0);

      // manhattan distance is very handy here
      if (rv.distance < 8) // 8 works nicely for speed << 70 and buggy gcfclient
        {
          socket.current_map = ob->map;
          socket.current_x   = ob->x;
          socket.current_y   = ob->y;

          socket_map_scroll (&socket, rv.distance_x, rv.distance_y);
          socket.floorbox_reset ();
          send_map_info (pl);
        }
    }

  if (socket.current_map != ob->map || socket.force_newmap)
    {
      clear_map (pl);
      socket.current_map = ob->map;
      send_map_info (pl);
    }
  else if (socket.current_x != ob->x || socket.current_y != ob->y)
    {
      int dx = ob->x - socket.current_x;
      int dy = ob->y - socket.current_y;

      socket_map_scroll (&socket, ob->x - socket.current_x, ob->y - socket.current_y);
      socket.floorbox_reset ();
    }

  socket.current_x = ob->x;
  socket.current_y = ob->y;
}

/**
 * RequestInfo is sort of a meta command. There is some specific
 * request of information, but we call other functions to provide
 * that information.
 */
void
RequestInfo (char *buf, int len, client *ns)
{
  char *params;

  /* find the first space, make it null, and update the
   * params pointer.
   */
  for (params = buf; *params; params++)
    if (*params == ' ')
      {
        *params++ = 0;
        break;
      }

  if (!strcmp (buf, "image_info"))
    send_image_info (ns, params);
  else if (!strcmp (buf, "image_sums"))
    send_image_sums (ns, params);
  else if (!strcmp (buf, "skill_info"))
    send_skill_info (ns, params);
  else if (!strcmp (buf, "spell_paths"))
    send_spell_paths (ns, params);
  else
    {
      // undo tokenisation above and send replyinfo with the request unchanged
      if (*params)
        *--params = ' ';

      ns->send_packet_printf ("replyinfo %s", buf);
    }
}

void
ExtCmd (char *buf, int len, player *pl)
{
  INVOKE_PLAYER (EXTCMD, pl, ARG_DATA (buf, len));
}

void
ExtiCmd (char *buf, int len, client *ns)
{
  INVOKE_CLIENT (EXTICMD, ns, ARG_DATA (buf, len));
}

void
client::mapinfo_queue_clear ()
{
  for (auto (i, mapinfo_queue.begin ()); i != mapinfo_queue.end (); ++i)
    free (*i);

  mapinfo_queue.clear ();
}

bool
client::mapinfo_try (char *buf)
{
  char *token = buf;
  buf += strlen (buf) + 9;

  // initial map and its origin
  maptile *map = pl->observe->map;
  int mapx = pl->ns->mapx / 2 - pl->observe->x;
  int mapy = pl->ns->mapy / 2 - pl->observe->y;
  int max_distance = 8;     // limit maximum path length to something generous

  while (*buf && map && max_distance)
    {
      int dir = *buf++ - '1';

      if (dir >= 0 && dir <= 3)
        {
          if (!map->tile_path [dir])
            map = 0;
          else if (map->tile_available (dir, false))
            {
              maptile *neigh = map->tile_map [dir];

              switch (dir)
                {
                  case 0: mapy -= neigh->height; break;
                  case 2: mapy += map  ->height; break;
                  case 3: mapx -= neigh->width ; break;
                  case 1: mapx += map  ->width ; break;
                }

              map = neigh;
              --max_distance;
            }
          else
            return 0;
        }
      else
        max_distance = 0;
    }

  if (!max_distance)
    send_packet_printf ("mapinfo %s error", token);
  else if (!map || !map->path)
    send_packet_printf ("mapinfo %s nomap", token);
  else
    {
      int flags = 0;

      if (map->tile_path[0]) flags |= 1;
      if (map->tile_path[1]) flags |= 2;
      if (map->tile_path[2]) flags |= 4;
      if (map->tile_path[3]) flags |= 8;

      send_packet_printf ("mapinfo %s spatial %d %d %d %d %d %s", token, flags, mapx, mapy, map->width, map->height, &map->path);
    }

  return 1;
}

void
client::mapinfo_queue_run ()
{
  if (mapinfo_queue.empty () || !pl)
    return;

  for (int i = 0; i < mapinfo_queue.size (); ++i)
    if (mapinfo_try (mapinfo_queue [i]))
      {
        free (mapinfo_queue [i]);
        mapinfo_queue.erase (i);
      }
    else
      ++i;
}

void
MapInfoCmd (char *buf, int len, player *pl)
{
  // <mapinfo tag spatial tile-path
  // >mapinfo tag spatial flags x y w h hash

  char *token = buf;

  if (!(buf = strchr (buf, ' ')))
    return;

  if (!strncmp (buf, " spatial ", 9))
    {
      char *copy = strdup (token);
      copy [buf - token] = 0;

#if 0
      // this makes only sense when we flush the buffer immediately
      if (pl->ns->mapinfo_try (copy))
        free (copy);
      else
#endif
        pl->ns->mapinfo_queue.push_back (copy);
    }
  else
    pl->ns->send_packet_printf ("mapinfo %s unsupported", token);
}

/** This is the Setup cmd */
void
SetUp (char *buf, int len, client * ns)
{
  INVOKE_CLIENT (SETUP, ns, ARG_DATA (buf, len));
}

/**
 * The client has requested to be added to the game.
 * This is what takes care of it.  We tell the client how things worked out.
 * I am not sure if this file is the best place for this function.  however,
 * it either has to be here or init_sockets needs to be exported.
 */
void
AddMeCmd (char *buf, int len, client *ns)
{
  INVOKE_CLIENT (ADDME, ns, ARG_DATA (buf, len));
}

/*
#define MSG_TYPE_BOOK            1
#define MSG_TYPE_CARD            2
#define MSG_TYPE_PAPER           3
#define MSG_TYPE_SIGN            4
#define MSG_TYPE_MONUMENT        5
#define MSG_TYPE_SCRIPTED_DIALOG 6*/

/** Reply to ExtendedInfos command */
void
ToggleExtendedText (char *buf, int len, client * ns)
{
  char cmdback[MAX_BUF];
  char temp[10];
  char command[50];
  int info, nextinfo, i, flag;

  cmdback[0] = '\0';

  nextinfo = 0;
  while (1)
    {
      /* 1. Extract an info */
      info = nextinfo;

      while ((info < len) && (buf[info] == ' '))
        info++;

      if (info >= len)
        break;

      nextinfo = info + 1;

      while ((nextinfo < len) && (buf[nextinfo] != ' '))
        nextinfo++;

      if (nextinfo - info >= 49)        /*Erroneous info asked */
        continue;

      strncpy (command, &(buf[info]), nextinfo - info);
      command[nextinfo - info] = '\0';
      /* 2. Interpret info */
      i = sscanf (command, "%d", &flag);

      if ((i == 1) && (flag > 0) && (flag <= MSG_TYPE_LAST))
        ns->supported_readables |= (1 << flag);
      /*3. Next info */
    }

  /* Send resulting state */
  strcpy (cmdback, "ExtendedTextSet");

  for (i = 0; i <= MSG_TYPE_LAST; i++)
    if (ns->supported_readables & (1 << i))
      {
        strcat (cmdback, " ");
        snprintf (temp, sizeof (temp), "%d", i);
        strcat (cmdback, temp);
      }

  ns->send_packet (cmdback);
}

/**
 * This handles the general commands from the client (ie, north, fire, cast,
 * etc.)
 */
void
PlayerCmd (char *buf, int len, player *pl)
{
  /* Check if there is a count.  In theory, a zero count could also be
   * sent, so check for that also.
   */
  if (atoi (buf) || buf[0] == '0')
    {
      pl->count = atoi ((char *) buf);

      buf = strchr (buf, ' ');  /* advance beyond the numbers */
      if (!buf)
        return;

      buf++;
    }

  execute_newserver_command (pl->ob, (char *) buf);

  /* Perhaps something better should be done with a left over count.
   * Cleaning up the input should probably be done first - all actions
   * for the command that issued the count should be done before any other
   * commands.
   */
  pl->count = 0;
}

/**
 * This handles the general commands from the client (ie, north, fire, cast,
 * etc.).  It is a lot like PlayerCmd above, but is called with the
 * 'ncom' method which gives more information back to the client so it
 * can throttle.
 */
void
NewPlayerCmd (char *buf, int len, player *pl)
{
  if (len <= 6)
    {
      LOG (llevDebug, "%s: corrupt ncom command <%s>: not long enough (%d) - discarding\n", pl->ns->host, buf, len);
      return;
    }

  uint16 cmdid  = net_uint16 ((uint8 *)buf);
  sint32 repeat = net_sint32 ((uint8 *)buf + 2);

  /* -1 is special - no repeat, but don't update */
  if (repeat != -1)
    pl->count = repeat;

  buf += 6; //len -= 6;

  execute_newserver_command (pl->ob, buf);

  /* Perhaps something better should be done with a left over count.
   * Cleaning up the input should probably be done first - all actions
   * for the command that issued the count should be done before any other
   * commands.
   */
  pl->count = 0;

  //TODO: schmorp thinks whatever this calculates, it makes no sense at all
  int time = fabs (pl->ob->speed) < 0.001
                ? time = MAX_TIME * 100
                : time = (int) (MAX_TIME / fabs (pl->ob->speed));

  /* Send confirmation of command execution now */
  packet sl ("comc");
  sl << uint16 (cmdid) << uint32 (time);
  pl->ns->send_packet (sl);
}

/** This is a reply to a previous query. */
void
ReplyCmd (char *buf, int len, client *ns)
{
  if (ns->state == ST_CUSTOM)
    {
      INVOKE_CLIENT (REPLY, ns, ARG_DATA (buf, len));
      return;
    }

  if (!ns->pl)
    return; //TODO: depends on the exact reply we are after
            //TODO: but right now, we always have a ns->pl

  player *pl = ns->pl;

  /* This is to synthesize how the data would be stored if it
   * was normally entered. A bit of a hack, and should be cleaned up
   * once all the X11 code is removed from the server.
   *
   * We pass 13 to many of the functions because this way they
   * think it was the carriage return that was entered, and the
   * function then does not try to do additional input.
   */
  snprintf (pl->write_buf, sizeof (pl->write_buf), ":%s", buf);

  /* this avoids any hacking here */

  switch (ns->state)
    {
      case ST_PLAYING:
        LOG (llevError, "Got reply message with ST_PLAYING input state\n");
        break;

      case ST_GET_PARTY_PASSWORD:    /* Get password for party */
        receive_party_password (pl->ob, 13);
        break;

      default:
        LOG (llevError, "Unknown input state: %d\n", ns->state);
    }
}

/**
 * Client tells its version.  If there is a mismatch, we close the
 * socket.  In real life, all we should care about is the client having
 * something older than the server.  If we assume the client will be
 * backwards compatible, having it be a later version should not be a 
 * problem.
 */
void
VersionCmd (char *buf, int len, client * ns)
{
  if (!buf)
    {
      LOG (llevError, "CS: received corrupted version command\n");
      return;
    }

  ns->cs_version = atoi (buf);
  ns->sc_version = ns->cs_version;

  LOG (llevDebug, "connection from client <%s>\n", buf);

  //TODO: should log here just for statistics

  //if (VERSION_CS != ns->cs_version)
  //  unchecked;

  char *cp = strchr (buf + 1, ' ');
  if (!cp)
    return;

  ns->sc_version = atoi (cp);

  //if (VERSION_SC != ns->sc_version)
  //  unchecked;

  cp = strchr (cp + 1, ' ');

  if (cp)
    {
      ns->version = cp + 1;

      if (ns->sc_version < 1026)
        ns->send_packet_printf ("drawinfo %d %s", NDI_RED,
                   "**** VERSION WARNING ****\n**** CLIENT IS TOO OLD!! UPDATE THE CLIENT!! ****");
    }
}

/** sound related functions. */
void
SetSound (char *buf, int len, client * ns)
{
  ns->sound = atoi (buf);
}

/** client wants the map resent */
void
MapRedrawCmd (char *buf, int len, player *pl)
{
/* This function is currently disabled; just clearing the map state results in
 * display errors. It should clear the cache and send a newmap command.
 * Unfortunately this solution does not work because some client versions send
 * a mapredraw command after receiving a newmap command.
 */
}

/**
 * Moves an object (typically, container to inventory).
 * syntax is: move (to) (tag) (nrof)
 */
void
MoveCmd (char *buf, int len, player *pl)
{
  int to, tag, nrof;

  if (3 != sscanf (buf, "%d %d %d", &to, &tag, &nrof))
    {
      LOG (llevError, "Incomplete move command: %s\n", buf);
      return;
    }

  esrv_move_object (pl->ob, to, tag, nrof);
}

/******************************************************************************
 *
 * Start of commands the server sends to the client.
 *
 ******************************************************************************/

/**
 * Asks the client to query the user.  This way, the client knows
 * it needs to send something back (vs just printing out a message)
 */
void
send_query (client *ns, uint8 flags, const char *text)
{
  ns->send_packet_printf ("query %d %s", flags, text ? text : "");
}

/**
 * Get player's current range attack in obuf.
 */
static void
rangetostring (player *pl, char *obuf)
{
  dynbuf_text buf;

  if (pl->ranged_ob)
    buf << " Range"  << (pl->ob->current_weapon == pl->ranged_ob ? "*" : "") << ": " << pl->ranged_ob->name;

  if (pl->combat_ob)
    buf << " Combat" << (pl->ob->current_weapon == pl->combat_ob ? "*" : "") << ": " << pl->combat_ob->name;

#if 0
  //TODO: remove this when slot system is working, this is only for debugging
  if (pl->ob->chosen_skill)
    buf << " Skill*: "  << pl->ob->chosen_skill->name;
#endif

  //TODO: maybe golem should become the current_weapon, quite simply?
  if (pl->golem)
    buf << " Golem*: "  << pl->golem->name;

  buf << '\0';
  buf.linearise (obuf);
}

#define AddIfInt64(Old,New,Type) if (Old != New) {\
                        Old = New; \
                        sl << uint8 (Type) << uint64 (New); \
                       }

#define AddIfInt(Old,New,Type) if (Old != New) {\
                        Old = New; \
                        sl << uint8 (Type) << uint32 (New); \
                       }

#define AddIfShort(Old,New,Type) if (Old != New) {\
                        Old = New; \
                        sl << uint8 (Type) << uint16 (New); \
                       }

#define AddIfFloat(Old,New,Type,mult) if (Old != New) {\
                        Old = New; \
                        sl << uint8 (Type) << uint32 (New*FLOAT_MULTI*mult); \
                        }

#define AddIfString(Old,New,Type) if (Old == NULL || strcmp(Old,New)) {\
                        free(Old); Old = strdup (New);\
                        sl << uint8 (Type) << data8 (New); \
                        }

/**
 * Sends a statistics update.  We look at the old values,
 * and only send what has changed.  Stat mapping values are in newclient.h
 * Since this gets sent a lot, this is actually one of the few binary
 * commands for now.
 */
void
esrv_update_stats (player *pl)
{
  char buf[MAX_BUF];
  uint16 flags;

  client *ns = pl->ns;
  if (!ns)
    return;

  object *ob = pl->observe;
  if (!ob)
    return;

  player *opl = ob->contr ? static_cast<player *>(ob->contr) : pl;

  packet sl ("stats");

  AddIfShort (ns->last_stats.hp,       ob->stats.hp,       CS_STAT_HP);
  AddIfShort (ns->last_stats.maxhp,    ob->stats.maxhp,    CS_STAT_MAXHP);
  AddIfShort (ns->last_stats.sp,       ob->stats.sp,       CS_STAT_SP);
  AddIfShort (ns->last_stats.maxsp,    ob->stats.maxsp,    CS_STAT_MAXSP);
  AddIfShort (ns->last_stats.grace,    ob->stats.grace,    CS_STAT_GRACE);
  AddIfShort (ns->last_stats.maxgrace, ob->stats.maxgrace, CS_STAT_MAXGRACE);
  AddIfShort (ns->last_stats.Str,      ob->stats.Str,      CS_STAT_STR);
  AddIfShort (ns->last_stats.Dex,      ob->stats.Dex,      CS_STAT_DEX);
  AddIfShort (ns->last_stats.Con,      ob->stats.Con,      CS_STAT_CON);
  AddIfShort (ns->last_stats.Int,      ob->stats.Int,      CS_STAT_INT);
  AddIfShort (ns->last_stats.Wis,      ob->stats.Wis,      CS_STAT_WIS);
  AddIfShort (ns->last_stats.Pow,      ob->stats.Pow,      CS_STAT_POW);
  AddIfShort (ns->last_stats.Cha,      ob->stats.Cha,      CS_STAT_CHA);

  for (int s = 0; s < NUM_SKILLS; s++)
    if (object *skill = opl->last_skill_ob [s])
      if (skill->stats.exp != ns->last_skill_exp [s])
        {
          ns->last_skill_exp [s] = skill->stats.exp;

          /* Always send along the level if exp changes.  This is only
           * 1 extra byte, but keeps processing simpler.
           */
          sl << uint8 (s + CS_STAT_SKILLINFO)
             << uint8 (skill->level)
             << uint64 (skill->stats.exp);
        }

  AddIfInt64 (ns->last_stats.exp,  ob->stats.exp, CS_STAT_EXP64);
  AddIfShort (ns->last_level,      ob->level, CS_STAT_LEVEL);
  AddIfShort (ns->last_stats.wc,   ob->stats.wc, CS_STAT_WC);
  AddIfShort (ns->last_stats.ac,   ob->stats.ac, CS_STAT_AC);
  AddIfShort (ns->last_stats.dam,  ob->stats.dam, CS_STAT_DAM);
  AddIfFloat (ns->last_speed,      ob->speed, CS_STAT_SPEED, 1.f/TICK);
  AddIfShort (ns->last_stats.food, ob->stats.food, CS_STAT_FOOD);
  AddIfFloat (ns->last_weapon_sp,  pl->weapon_sp, CS_STAT_WEAP_SP, 1.f/TICK);
  AddIfInt   (ns->last_weight_limit, weight_limit[ob->stats.Str], CS_STAT_WEIGHT_LIM);

  flags = 0;

  if (opl->fire_on)
    flags |= SF_FIREON;

  if (opl->run_on)
    flags |= SF_RUNON;

  AddIfShort (ns->last_flags, flags, CS_STAT_FLAGS);

  if (ns->sc_version < 1025)
    { AddIfShort (ns->last_resist[ATNR_PHYSICAL], ob->resist[ATNR_PHYSICAL], CS_STAT_ARMOUR) }
  else
    for (int i = 0; i < NROFATTACKS; i++)
      {
        /* Skip ones we won't send */
        if (atnr_cs_stat[i] == -1)
          continue;

        AddIfShort (ns->last_resist[i], ob->resist[i], atnr_cs_stat[i]);
      }

  if (pl->ns->monitor_spells)
    {
      AddIfInt (ns->last_path_attuned,  ob->path_attuned,  CS_STAT_SPELL_ATTUNE);
      AddIfInt (ns->last_path_repelled, ob->path_repelled, CS_STAT_SPELL_REPEL);
      AddIfInt (ns->last_path_denied,   ob->path_denied,   CS_STAT_SPELL_DENY);
    }

  rangetostring (opl, buf);  /* we want use the new fire & run system in new client */
  AddIfString (ns->stats.range, buf, CS_STAT_RANGE);
  set_title (ob, buf);
  AddIfString (ns->stats.title, buf, CS_STAT_TITLE);

  /* Only send it away if we have some actual data */
  if (sl.length () > 6)
    ns->send_packet (sl);
}

/**
 * Tells the client that here is a player it should start using.
 */
void
esrv_new_player (player *pl)
{
  sint32 weight = pl->ob->client_weight ();

  packet sl ("player");

  sl << uint32 (pl->ob->count)
     << uint32 (weight)
     << uint32 (pl->ob->face)
     << data8  (pl->ob->name);

  pl->ns->last_weight = weight;
  pl->ns->send_packet (sl);
}

/******************************************************************************
 *
 * Start of map related commands.
 *
 ******************************************************************************/

/** Clears a map cell */
static void
map_clearcell (struct MapCell *cell, int count)
{
  cell->faces[0]  = 0;
  cell->faces[1]  = 0;
  cell->faces[2]  = 0;
  cell->smooth[0] = 0;
  cell->smooth[1] = 0;
  cell->smooth[2] = 0;
  cell->count     = count;
  cell->stat_hp   = 0;
  cell->flags     = 0;
  cell->player    = 0;
}

#define MAX_LAYERS	3

/**
 * Removes the need to replicate the same code for each layer.
 * this returns true if this space is now in fact different than
 * it was.
 * sl is the socklist this data is going into.
 * ns is the socket we are working on - all the info we care
 * about is in this socket structure, so now need not pass the
 * entire player object.
 * layer is the layer to update, with 2 being the floor and 0 the
 * top layer (this matches what the GET_MAP_FACE and GET_MAP_FACE_OBJ)
 * take.  Interesting to note that before this function, the map1 function
 * numbers the spaces differently - I think this was a leftover from
 * the map command, where the faces stack up.  Sinces that is no longer
 * the case, it seems to make more sense to have these layer values
 * actually match.
 */
static int
update_space (packet &sl, client &ns, mapspace &ms, MapCell &lastcell, int layer)
{
  object *ob = ms.faces_obj [layer];

  /* If there is no object for this space, or if the face for the object
   * is the blank face, set the face number to zero.
   * else if we have the stored head object for this space, that takes
   * precedence over the other object for this space.
   * otherwise, we do special head processing
   */
  uint16 face_num = ob && ob->face != blank_face ? ob->face : 0;

  /* We've gotten what face we want to use for the object.  Now see if
   * if it has changed since we last sent it to the client.
   */
  if (lastcell.faces[layer] != face_num)
    {
      lastcell.faces[layer] = face_num;

      if (!ns.faces_sent[face_num])
        if (ob)
          ns.send_faces (ob);
        else
          ns.send_face (face_num, 10);

      sl << uint16 (face_num);
      return 1;
    }

  /* Nothing changed */
  return 0;
}

// prefetch (and touch) all maps within a specific distancd
static void
prefetch_surrounding_maps (maptile *map, int distance)
{
  map->touch ();

  if (--distance)
    for (int dir = 4; --dir; )
      if (const shstr &path = map->tile_path [dir])
        if (maptile *&neigh = map->tile_map [dir])
          prefetch_surrounding_maps (neigh, distance);
        else
          neigh = maptile::find_async (path, map);
}

// prefetch a generous area around the player
static void
prefetch_surrounding_maps (object *op)
{
  prefetch_surrounding_maps (op->map, 3);
}

/**
 * Draws client map.
 */
void
draw_client_map (player *pl)
{
  object *ob = pl->observe;
  if (!ob->active)
    return;

  /* If player is just joining the game, he isn't here yet, so the map
   * can get swapped out.  If so, don't try to send them a map.  All will
   * be OK once they really log in.
   */
  if (!ob->map || ob->map->in_memory != MAP_ACTIVE)
    return;

  int startlen, oldlen;

  check_map_change (pl);
  prefetch_surrounding_maps (pl->ob);

  /* do LOS after calls to update_position */
  /* unfortunately, we need to udpate los when observing, currently */
  if (pl->do_los || pl->observe != pl->ob)
    {
      pl->do_los = 0;
      pl->update_los ();
    }

  /**
   * This function uses the new map1 protocol command to send the map
   * to the client.  It is necessary because the old map command supports
   * a maximum map size of 15x15.
   * This function is much simpler than the old one.  This is because
   * the old function optimized to send as few face identifiers as possible,
   * at the expense of sending more coordinate location (coordinates were
   * only 1 byte, faces 2 bytes, so this was a worthwhile savings).  Since
   * we need 2 bytes for coordinates and 2 bytes for faces, such a trade off
   * maps no sense.  Instead, we actually really only use 12 bits for coordinates,
   * and use the other 4 bits for other informatiion.   For full documentation
   * of what we send, see the doc/Protocol file.
   * I will describe internally what we do:
   * the ns->lastmap shows how the map last looked when sent to the client.
   * in the lastmap structure, there is a cells array, which is set to the
   * maximum viewable size (As set in config.h).
   * in the cells, there are faces and a count value.
   * we use the count value to hold the darkness value.  If -1, then this space
   *   is not viewable.
   * we use faces[0] faces[1] faces[2] to hold what the three layers
   * look like.
   */

  client &socket = *pl->ns;

  packet sl (socket.mapmode == Map1Cmd ? "map1" : "map1a");

  startlen = sl.length ();

  int hx = socket.mapx / 2;
  int hy = socket.mapy / 2;

  ordered_mapwalk_begin (ob, -hx, -hy, hx, hy)
    int ax = dx + hx;
    int ay = dy + hy;

    int mask = (ax << 10) | (ay << 4);
    MapCell &lastcell = socket.lastmap.cells[ax][ay];

    /* If the coordinates are not valid, or it is too dark to see,
     * we tell the client as such
     */
    if (!m)
      {
        /* space is out of map. Update space and clear values
         * if this hasn't already been done. If the space is out
         * of the map, it shouldn't have a head.
         */
        if (lastcell.count != -1)
          {
            sl << uint16 (mask);
            map_clearcell (&lastcell, -1);
          }

        continue;
      }
      
    int d = pl->blocked_los_uc (dx, dy);

    if (d > 3)
      {
        /* This block deals with spaces that are not visible for whatever
         * reason. Still may need to send the head for this space.
         */
        if (lastcell.count != -1
            || lastcell.faces[0] 
            || lastcell.faces[1] 
            || lastcell.faces[2]
            || lastcell.stat_hp 
            || lastcell.flags 
            || lastcell.player)
          sl << uint16 (mask);

        /* properly clear a previously sent big face */
        map_clearcell (&lastcell, -1);
      }
    else
      {
        /* In this block, the space is visible.
         */

        /* Rather than try to figure out what everything that we might
         * need to send is, then form the packet after that,
         * we presume that we will in fact form a packet, and update
         * the bits by what we do actually send.  If we send nothing,
         * we just back out sl.length () to the old value, and no harm
         * is done.
         * I think this is simpler than doing a bunch of checks to see
         * what if anything we need to send, setting the bits, then
         * doing those checks again to add the real data.
         */
        oldlen  = sl.length ();

        sl << uint16 (mask);

        unsigned char dummy;
        unsigned char *last_ext = &dummy;

        /* Darkness changed */
        if (lastcell.count != d)
          {
            mask |= 0x8;

            *last_ext |= 0x80;
            last_ext = &sl[sl.length ()];
            sl << uint8 (d);
          }

        lastcell.count = d;

        mapspace &ms = m->at (nx, ny);
        ms.update ();

        // extmap handling
        uint8 stat_hp    = 0;
        uint8 stat_width = 0;
        uint8 flags      = 0;
        tag_t player     = 0;

        // send hp information, if applicable
        if (object *op = ms.faces_obj [0])
          if (op->is_head () && !op->invisible)
            {
              if (op->stats.maxhp > op->stats.hp
                  && op->stats.maxhp > 0
                  && (op->type == PLAYER
                      || op->type == DOOR // does not work, have maxhp 0
                      || QUERY_FLAG (op, FLAG_MONSTER)
                      || QUERY_FLAG (op, FLAG_ALIVE)
                      || QUERY_FLAG (op, FLAG_GENERATOR)))
                {
                  stat_hp = 255 - (op->stats.hp * 255 + 254) / op->stats.maxhp;
                  stat_width = op->arch->max_x - op->arch->x; //TODO: should be upper-left edge
                }

              if (expect_false (op->has_dialogue ()))
                flags |= 1;

              if (expect_false (op->type == PLAYER))
                player = op == ob     ? pl->ob->count
                       : op == pl->ob ? ob->count
                       :                op->count;
            }

        if (expect_false (lastcell.stat_hp != stat_hp))
          {
            lastcell.stat_hp = stat_hp;

            mask |= 0x8;
            *last_ext |= 0x80;
            last_ext = &sl[sl.length ()];

            sl << uint8 (5) << uint8 (stat_hp);

            if (stat_width > 1)
              {
                *last_ext |= 0x80;
                last_ext = &sl[sl.length ()];

                sl << uint8 (6) << uint8 (stat_width);
              }
          }

        if (expect_false (lastcell.player != player))
          {
            lastcell.player = player;

            mask |= 0x8;
            *last_ext |= 0x80;
            last_ext = &sl[sl.length ()];

            sl << uint8 (0x47) << uint8 (4) << (uint32)player;
          }

        if (expect_false (lastcell.flags != flags))
          {
            lastcell.flags = flags;

            mask |= 0x8;
            *last_ext |= 0x80;
            last_ext = &sl[sl.length ()];

            sl << uint8 (8) << uint8 (flags);
          }

        // faces

        /* Floor face */
        if (update_space (sl, socket, ms, lastcell, 2))
          mask |= 0x4;

        /* Middle face */
        if (update_space (sl, socket, ms, lastcell, 1))
          mask |= 0x2;

        if (expect_false (ob->invisible)
            && ob->invisible & (ob->invisible < 50 ? 1 : 7)
            && ms.player () == ob)
          {
            // force player to be visible to himself if invisible
            if (lastcell.faces[0] != ob->face)
              {
                lastcell.faces[0] = ob->face;

                mask |= 0x1;
                sl << uint16 (ob->face);

                socket.send_faces (ob);
              }
          }
        /* Top face */
        else if (update_space (sl, socket, ms, lastcell, 0))
          mask |= 0x1;

        /* Check to see if we are in fact sending anything for this
         * space by checking the mask.  If so, update the mask.
         * if not, reset the len to that from before adding the mask
         * value, so we don't send those bits.
         */
        if (mask & 0xf)
          sl[oldlen + 1] = mask & 0xff;
        else
          sl.reset (oldlen);
      }                   /* else this is a viewable space */
  ordered_mapwalk_end

  socket.flush_fx ();

  if (sl.length () > startlen || socket.sent_scroll)
    {
      socket.send_packet (sl);
      socket.sent_scroll = 0;
    }
}

/*****************************************************************************/
/* GROS: The following one is used to allow a plugin to send a generic cmd to*/
/* a player. Of course, the client need to know the command to be able to    */
/* manage it !                                                               */
/*****************************************************************************/
void
send_plugin_custom_message (object *pl, char *buf)
{
  pl->contr->ns->send_packet (buf);
}

/**
 * This sends the skill number to name mapping.  We ignore
 * the params - we always send the same info no matter what.
 */
void
send_skill_info (client *ns, char *params)
{
  packet sl;
  sl << "replyinfo skill_info\n";

  for (int i = 1; i < NUM_SKILLS; i++)
    sl.printf ("%d:%s\n", i + CS_STAT_SKILLINFO, &skill_names [i]);

  if (sl.length () > MAXSOCKBUF)
    {
      LOG (llevError, "Buffer overflow in send_skill_info!\n");
      fatal (0);
    }

  ns->send_packet (sl);
}

/**
 * This sends the spell path to name mapping.  We ignore
 * the params - we always send the same info no matter what.
 */
void
send_spell_paths (client * ns, char *params)
{
  packet sl;

  sl << "replyinfo spell_paths\n";

  for (int i = 0; i < NRSPELLPATHS; i++)
    sl.printf ("%d:%s\n", 1 << i, spellpathnames[i]);

  if (sl.length () > MAXSOCKBUF)
    {
      LOG (llevError, "Buffer overflow in send_spell_paths!\n");
      fatal (0);
    }

  ns->send_packet (sl);
}

/**
 * This looks for any spells the player may have that have changed their stats.
 * it then sends an updspell packet for each spell that has changed in this way
 */
void
esrv_update_spells (player *pl)
{
  if (!pl->ns)
    return;

  if (!pl->ns->monitor_spells)
    return;

  for (object *spell = pl->ob->inv; spell; spell = spell->below)
    {
      if (spell->type == SPELL)
        {
          int flags = 0;

          /* check if we need to update it */
          if (spell->cached_sp != SP_level_spellpoint_cost (pl->ob, spell, SPELL_MANA))
            {
              spell->cached_sp = SP_level_spellpoint_cost (pl->ob, spell, SPELL_MANA);
              flags |= UPD_SP_MANA;
            }

          if (spell->cached_grace != SP_level_spellpoint_cost (pl->ob, spell, SPELL_GRACE))
            {
              spell->cached_grace = SP_level_spellpoint_cost (pl->ob, spell, SPELL_GRACE);
              flags |= UPD_SP_GRACE;
            }

          if (spell->cached_eat != spell->stats.dam + SP_level_dam_adjust (pl->ob, spell))
            {
              spell->cached_eat = spell->stats.dam + SP_level_dam_adjust (pl->ob, spell);
              flags |= UPD_SP_DAMAGE;
            }

          if (flags)
            {
              packet sl;

              sl << "updspell "
                 << uint8 (flags)
                 << uint32 (spell->count);

              if (flags & UPD_SP_MANA  ) sl << uint16 (spell->cached_sp);
              if (flags & UPD_SP_GRACE ) sl << uint16 (spell->cached_grace);
              if (flags & UPD_SP_DAMAGE) sl << uint16 (spell->cached_eat);

              pl->ns->send_packet (sl);
            }
        }
    }
}

void
esrv_remove_spell (player *pl, object *spell)
{
  if (!pl->ns->monitor_spells)
    return;

  if (!pl || !spell || spell->env != pl->ob)
    {
      LOG (llevError, "Invalid call to esrv_remove_spell");
      return;
    }

  packet sl ("delspell");

  sl << uint32 (spell->count);

  pl->ns->send_packet (sl);
}

/* appends the spell *spell to the Socklist we will send the data to. */
static void
append_spell (player *pl, packet &sl, object *spell)
{
  int i, skill = 0;

  if (!(spell->name))
    {
      LOG (llevError, "item number %d is a spell with no name.\n", spell->count);
      return;
    }

  /* store costs and damage in the object struct, to compare to later */
  spell->cached_sp    = SP_level_spellpoint_cost (pl->ob, spell, SPELL_MANA);
  spell->cached_grace = SP_level_spellpoint_cost (pl->ob, spell, SPELL_GRACE);
  spell->cached_eat   = spell->stats.dam + SP_level_dam_adjust (pl->ob, spell);

  /* figure out which skill it uses, if it uses one */
  if (spell->skill)
    if (object *tmp = pl->find_skill (spell->skill))
      skill = tmp->subtype + CS_STAT_SKILLINFO;

  // spells better have a face
  if (!spell->face)
    {
      LOG (llevError, "%s: spell has no face, but face is mandatory.\n", &spell->name);
      spell->face = face_find ("burnout.x11", blank_face);
    }

  pl->ns->send_face (spell->face);

  /* send the current values */
  sl << uint32 (spell->count)
     << uint16 (spell->level)
     << uint16 (spell->casting_time)
     << uint16 (spell->cached_sp)
     << uint16 (spell->cached_grace)
     << uint16 (spell->cached_eat)
     << uint8  (skill)
     << uint32 (spell->path_attuned)
     << uint32 (spell->face)
     << data8  (spell->name)
     << data16 (spell->msg);
}

/**
 * This tells the client to add the spell *ob, if *ob is NULL, then add 
 * all spells in the player's inventory.
 */
void
esrv_add_spells (player *pl, object *spell)
{
  if (!pl)
    {
      LOG (llevError, "esrv_add_spells, tried to add a spell to a NULL player");
      return;
    }

  if (!pl->ns->monitor_spells)
    return;

  packet sl ("addspell");

  if (!spell)
    {
      for (spell = pl->ob->inv; spell; spell = spell->below)
        {
          /* were we to simply keep appending data here, we could exceed 
           * MAXSOCKBUF if the player has enough spells to add, we know that
           * append_spells will always append 19 data bytes, plus 4 length
           * bytes and 3 strings (because that is the spec) so we need to
           * check that the length of those 3 strings, plus the 23 bytes, 
           * won't take us over the length limit for the socket, if it does,
           * we need to send what we already have, and restart packet formation 
           */
          if (spell->type != SPELL)
            continue;

          /* Seeing crashes by overflowed buffers.  Quick arithemetic seems
           * to show add_spell is 26 bytes + 2 strings.  However, the overun
           * is hundreds of bytes off, so correcting 22 vs 26 doesn't seem
           * like it will fix this
           */
          if (sl.length () > (MAXSOCKBUF - (26 + strlen (spell->name) + (spell->msg ? strlen (spell->msg) : 0))))
            {
              pl->ns->flush_fx ();
              pl->ns->send_packet (sl);

              sl.reset ();
              sl << "addspell ";
            }

          append_spell (pl, sl, spell);
        }
    }
  else if (spell->type != SPELL)
    {
      LOG (llevError, "Asked to send a non-spell object as a spell");
      return;
    }
  else
    append_spell (pl, sl, spell);

  if (sl.length () > MAXSOCKBUF)
    {
      LOG (llevError, "Buffer overflow in esrv_add_spells!\n");
      fatal (0);
    }

  /* finally, we can send the packet */
  pl->ns->flush_fx ();
  pl->ns->send_packet (sl);
}

